import lnschema_core as schema
import sqlmodel as sqm
from lamin_logger import logger
from lndb_setup import settings
from lndb_setup._hub import connect_hub
from supabase import Client

from ._entities import Entities
from ._schema import check_schema_version


def get_hub_with_authentication():
    hub = connect_hub()
    session = hub.auth.sign_in(
        email=settings.user.email, password=settings.user.password
    )
    hub.postgrest.auth(session.access_token)
    check_schema_version(hub)
    return hub


class hub:
    """Access the hub."""

    @classmethod
    def _load_entity(cls, entity):
        # we need to do this right now as we can't do a cyclic import
        # we first need to import all of lamindb (including load)
        # then lndb_hub is imported
        # once it's imported load can be imported here
        from lamindb.db._query import query

        return query.table_as_df(entity)

    @classmethod
    def push_instance(cls):
        """Push instance with all dobjects."""
        hub = get_hub_with_authentication()
        dobject_df = cls._load_entity("dobject")
        for id, v in dobject_df.index:
            cls.push_dobject(id, v, hub)

    @classmethod
    def push_dobject(cls, id: str, v: str = "1", hub: Client = None):
        """Push a single dobject."""
        if hub is None:
            hub = get_hub_with_authentication()
        entities = Entities(hub)
        with sqm.Session(settings.instance.db_engine()) as session:
            storage = session.exec(
                sqm.select(schema.storage).where(
                    schema.storage.root == str(settings.instance.storage_dir)
                )
            ).first()
            assert storage
            cls.__push_storage(entities, storage)

            instance = entities.instance.get_or_create(storage.id)

            dobject = session.get(schema.dobject, (id, v))
            assert dobject

            dtransform = session.get(schema.dtransform, dobject.dtransform_id)
            assert dtransform

            dtransform_ins = session.exec(
                sqm.select(schema.dtransform_in)
                .where(schema.dtransform_in.dobject_id == dobject.id)
                .where(schema.dtransform_in.dobject_v == dobject.v)
            ).all()

            jupynb = session.get(
                schema.jupynb, (dtransform.jupynb_id, dtransform.jupynb_v)
            )

            pipeline_run = session.get(schema.pipeline_run, dtransform.pipeline_run_id)

            usages = session.exec(
                sqm.select(schema.usage)
                .where(schema.usage.dobject_id == dobject.id)
                .where(schema.usage.dobject_v == dobject.v)
            ).all()

        assert jupynb or pipeline_run
        if jupynb:
            cls.__push_jupynb(entities, jupynb, instance["id"])
        if pipeline_run:
            pipeline = session.get(
                schema.pipeline, (pipeline_run.pipeline_id, pipeline_run.pipeline_v)
            )
            assert pipeline
            cls.__push_pipeline(entities, pipeline, instance["id"])
            cls.__push_pipeline_run(entities, pipeline_run, instance["id"])
        cls.__push_dtransform(entities, dtransform, instance["id"])
        cls.__push_dobject(entities, dobject, instance["id"])
        for dtransform_in in dtransform_ins:
            cls.__push_dtransform_in(entities, dtransform_in, instance["id"])
        for usage in usages:
            cls.__push_usage(entities, usage, instance["id"])

    @classmethod
    def delete_instance(cls):
        """Delete the instance on the hub."""
        hub = get_hub_with_authentication()
        entities = Entities(hub)
        instance = entities.instance.get()
        dobject_df = cls._load_entity("dobject")
        usage_df = cls._load_entity("usage")
        dtransform_in_df = cls._load_entity("dtransform_in")
        jupynb_df = cls._load_entity("jupynb")
        dtransform_df = cls._load_entity("dtransform")
        pipeline_df = cls._load_entity("pipeline")
        pipeline_run_df = cls._load_entity("pipeline_run")
        storage_df = cls._load_entity("storage")
        for id, v in dobject_df.index:
            logger.info(f"Delete dobject ({id}, {v}).")
            for dtransform_id in dtransform_in_df[
                (dtransform_in_df["dobject_id"] == id)
                & (dtransform_in_df["dobject_v"] == v)  # noqa
            ]["dtransform_id"].values:
                entities.dtransform_in.delete(dtransform_id, id, v)
            for usage_id in usage_df[
                (usage_df["dobject_id"] == id) & (usage_df["dobject_v"] == v)
            ].index:
                entities.usage.delete(usage_id)
            entities.dobject.delete(id, v, instance["id"])
        for id in dtransform_df.index:
            logger.info(f"Delete dtransform ({id}).")
            entities.dtransform.delete(id, instance["id"])
        for id, v in jupynb_df.index:
            logger.info(f"Delete jupynb ({id}, {v}).")
            entities.jupynb.delete(id, v, instance["id"])
        for id in pipeline_run_df.index:
            logger.info(f"Delete pipeline run ({id}).")
            entities.pipeline_run.delete(id, instance["id"])
        for id, v in pipeline_df.index:
            logger.info(f"Delete pipeline ({id}, {v}).")
            entities.pipeline.delete(id, v, instance["id"])
        entities.user_instance.delete(instance["id"])
        if entities.instance.user_count(instance["id"]) == 0:
            entities.instance.delete()
        for id in storage_df.index:
            logger.info(f"Delete storage ({id}).")
            if entities.storage.instance_count(id) == 0:
                entities.storage.delete(id)

    @classmethod
    def __push_storage(cls, entities: Entities, storage: schema.storage):
        if not entities.storage.exists(storage.root):
            logger.info(f"Push storage ({storage.root}).")
            entities.storage.insert(
                storage.id,
                storage.root,
                storage.region,
                storage.type,
            )
        else:
            logger.warning(
                f"storage ({storage.root}) already exists and is not pushed."
            )

    @classmethod
    def __push_dtransform(
        cls, entities: Entities, dtransform: schema.dtransform, instance_id: str
    ):
        if not entities.dtransform.exists(dtransform.id, instance_id):
            logger.info(f"Push dtransform ({dtransform.id}).")
            entities.dtransform.insert(
                dtransform.id, dtransform.jupynb_id, dtransform.jupynb_v, instance_id
            )
        else:
            logger.warning(
                f"dtransform ({dtransform.id}) already exists and is not pushed."
            )

    @classmethod
    def __push_jupynb(cls, entities: Entities, jupynb: schema.jupynb, instance_id: str):
        if not entities.jupynb.exists(jupynb.id, jupynb.v, instance_id):
            logger.info(f"Push jupynb ({jupynb.id}, {jupynb.v}).")
            entities.jupynb.insert(
                jupynb.id,
                jupynb.v,
                jupynb.name,
                instance_id,
                jupynb.time_created,
                jupynb.time_updated,
            )
        else:
            logger.warning(
                f"jupynb ({jupynb.id}, {jupynb.v}) already exists and is not pushed."
            )

    @classmethod
    def __push_pipeline(
        cls, entities: Entities, pipeline: schema.pipeline, instance_id: str
    ):
        if not entities.pipeline.exists(pipeline.id, pipeline.v, instance_id):
            logger.info(f"Push pipeline ({pipeline.id}, {pipeline.v}).")
            entities.pipeline.insert(
                pipeline.id,
                pipeline.v,
                pipeline.name,
                instance_id,
                pipeline.reference,
                pipeline.time_created,
            )
        else:
            logger.warning(
                f"pipeline ({pipeline.id}, {pipeline.v}) already exists and is not"
                " pushed."
            )

    @classmethod
    def __push_pipeline_run(
        cls, entities: Entities, pipeline_run: schema.pipeline_run, instance_id: str
    ):
        if not entities.pipeline_run.exists(pipeline_run.id, instance_id):
            logger.info(f"Push pipeline_run ({pipeline_run.id}).")
            entities.pipeline_run.insert(
                pipeline_run.id,
                pipeline_run.pipeline_id,
                pipeline_run.pipeline_v,
                pipeline_run.name,
                instance_id,
                pipeline_run.time_created,
            )
        else:
            logger.warning(
                f"pipeline ({pipeline_run.id}) already exists and is not pushed."
            )

    @classmethod
    def __push_dobject(
        cls, entities: Entities, dobject: schema.dobject, instance_id: str
    ):
        if not entities.dobject.exists(dobject.id, dobject.v, instance_id):
            logger.info(f"Push dobject ({dobject.id}, {dobject.v}).")
            entities.dobject.insert(
                dobject.id,
                dobject.v,
                dobject.name,
                dobject.file_suffix,
                dobject.dtransform_id,
                dobject.storage_id,
                instance_id,
                dobject.time_created,
                dobject.time_updated,
            )
        else:
            logger.warning(
                f"dobject ({dobject.id}, {dobject.v}) already exists and is not pushed."
            )

    @classmethod
    def __push_dtransform_in(
        cls, entities: Entities, dtransform_in: schema.dtransform_in, instance_id: str
    ):
        if not entities.dtransform_in.exists(
            dtransform_in.dtransform_id,
            dtransform_in.dobject_id,
            dtransform_in.dobject_v,
        ):
            entities.dtransform_in.insert(
                dtransform_in.dtransform_id,
                dtransform_in.dobject_id,
                dtransform_in.dobject_v,
                instance_id,
            )

    @classmethod
    def __push_usage(cls, entities: Entities, usage: schema.usage, instance_id: str):
        if not entities.usage.exists(usage.id):
            entities.usage.insert(
                usage.id,
                usage.type,
                instance_id,
                usage.time,
                usage.dobject_id,
                usage.dobject_v,
            )



import uuid
from datetime import datetime
from pathlib import Path
from typing import Union

from cloudpathlib import CloudPath
from lamin_logger import logger
from lndb_setup import settings
from supabase import Client


class Entities:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.instance = Instance(self.hub)
        self.user_instance = UserInstance(self.hub)
        self.dtransform = Dtransform(self.hub)
        self.pipeline = Pipeline(self.hub)
        self.jupynb = Jupynb(self.hub)
        self.pipeline_run = PipelineRun(self.hub)
        self.dtransform_in = DtransformIn(self.hub)
        self.dobject = Dobject(self.hub)
        self.storage = Storage(self.hub)
        self.usage = Usage(self.hub)


class Entity:
    def __init__(self, hub: Client, table_name: str, primary_keys: list) -> None:
        self.hub = hub
        self.table_name = table_name
        self.primary_keys = primary_keys

    def exists(self, primary_keys_values: dict):
        statement = self.hub.table(self.table_name).select("*")
        for key in self.primary_keys:
            statement = statement.eq(key, primary_keys_values[key])
        data = statement.execute()
        if len(data.data) > 0:
            return True
        return False

    def delete(self, primary_keys_values: dict):
        statement = self.hub.table(self.table_name).delete()
        for key in self.primary_keys:
            statement = statement.eq(key, primary_keys_values[key])
        statement.execute()

    def insert(self, fields: dict):
        data = self.hub.table(self.table_name).insert(fields).execute()
        assert len(data.data) == 1
        return data.data[0]


class Instance:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.instance_entity = Entity(hub, "instance", ["id"])
        self.user_instance_entity = Entity(
            hub, "user_instance", ["user_id", "instance_id"]
        )

    def get_or_create(self, storage_id: str):
        instance = self.get()
        if instance is None:
            instance = self.insert(storage_id)
            logger.info("Pushing instance.")
        if not self.user_instance_entity.exists(
            {
                "instance_id": instance["id"],
                "user_id": self.hub.auth.session().user.id.hex,
            }
        ):
            self.user_instance_entity.insert(
                {
                    "instance_id": instance["id"],
                    "user_id": self.hub.auth.session().user.id.hex,
                }
            )
        return instance

    def get(self):
        data = (
            self.hub.table("instance")
            .select("*")
            .eq("name", settings.instance.name)
            .execute()
        )
        if len(data.data) > 0:
            return data.data[0]
        return None

    def delete(self):
        (
            self.hub.table("instance")
            .delete()
            .eq("name", settings.instance.name)
            .execute()
        )

    def insert(self, storage_id: str):
        return self.instance_entity.insert(
            {
                "id": str(uuid.uuid4()),
                "name": settings.instance.name,
                "owner_id": self.hub.auth.session().user.id.hex,
                "storage_id": storage_id,
                "dbconfig": settings.instance._dbconfig,
                "cache_dir": str(settings.instance.cache_dir),
                "sqlite_file": str(settings.instance._sqlite_file),
                "sqlite_file_local": str(settings.instance._sqlite_file_local),
                "db": settings.instance.db,
            }
        )

    def user_count(self, instance_id: str):
        statement = (
            self.hub.table("user_instance").select("*").eq("instance_id", instance_id)
        )
        data = statement.execute()
        return len(data.data)


class UserInstance:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.user_instance_entity = Entity(
            hub, "user_instance", ["user_id", "instance_id"]
        )

    def exists(self, instance_id: str):
        return self.user_instance_entity.exists(
            {"user_id": self.hub.auth.session().user.id.hex, "instance_id": instance_id}
        )

    def delete(self, instance_id: str):
        self.user_instance_entity.delete(
            {"user_id": self.hub.auth.session().user.id.hex, "instance_id": instance_id}
        )

    def insert(self, instance_id: str):
        return self.user_instance_entity.delete(
            {"user_id": self.hub.auth.session().user.id.hex, "instance_id": instance_id}
        )


class Jupynb:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.jupynb_entity = Entity(hub, "jupynb", ["id", "v", "instance_id"])

    def exists(self, id: str, v: str, instance_id: str):
        return self.jupynb_entity.exists({"id": id, "v": v, "instance_id": instance_id})

    def delete(self, id: str, v: str, instance_id: str):
        self.jupynb_entity.delete({"id": id, "v": v, "instance_id": instance_id})

    def insert(
        self,
        id: str,
        v: str,
        name: str,
        instance_id: str,
        time_created: datetime,
        time_updated: datetime,
    ):
        return self.jupynb_entity.insert(
            {
                "id": id,
                "v": v,
                "name": name,
                "instance_id": instance_id,
                "owner_id": self.hub.auth.session().user.id.hex,
                "time_created": time_created.strftime("%Y-%m-%d %H:%M:%S"),
                "time_updated": time_updated.strftime("%Y-%m-%d %H:%M:%S"),
            }
        )


class Pipeline:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.pipeline_entity = Entity(hub, "pipeline", ["id", "v", "instance_id"])

    def exists(self, id: str, v: str, instance_id: str):
        return self.pipeline_entity.exists(
            {"id": id, "v": v, "instance_id": instance_id}
        )

    def delete(self, id: str, v: str, instance_id: str):
        self.pipeline_entity.delete({"id": id, "v": v, "instance_id": instance_id})

    def insert(
        self,
        id: str,
        v: str,
        name: str,
        instance_id: str,
        reference: str,
        time_created: datetime,
    ):
        return self.pipeline_entity.insert(
            {
                "id": id,
                "v": v,
                "name": name,
                "instance_id": instance_id,
                "owner_id": self.hub.auth.session().user.id.hex,
                "reference": reference,
                "time_created": time_created.strftime("%Y-%m-%d %H:%M:%S"),
            }
        )


class PipelineRun:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.pipeline_run_entity = Entity(hub, "pipeline_run", ["id", "instance_id"])

    def exists(self, id: str, instance_id: str):
        return self.pipeline_run_entity.exists({"id": id, "instance_id": instance_id})

    def delete(self, id: str, instance_id: str):
        self.pipeline_run_entity.delete({"id": id, "instance_id": instance_id})

    def insert(
        self,
        id: str,
        pipeline_id: str,
        pipeline_v: str,
        name: str,
        instance_id: str,
        time_created: datetime,
    ):
        return self.pipeline_run_entity.insert(
            {
                "id": id,
                "pipeline_id": pipeline_id,
                "pipeline_v": pipeline_v,
                "name": name,
                "instance_id": instance_id,
                "owner_id": self.hub.auth.session().user.id.hex,
                "time_created": time_created.strftime("%Y-%m-%d %H:%M:%S"),
            }
        )


class Dobject:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.dobject_entity = Entity(hub, "dobject", ["id", "v", "instance_id"])

    def exists(self, id: str, v: str, instance_id: str):
        return self.dobject_entity.exists(
            {"id": id, "v": v, "instance_id": instance_id}
        )

    def delete(self, id: str, v: str, instance_id: str):
        self.dobject_entity.delete({"id": id, "v": v, "instance_id": instance_id})

    def insert(
        self,
        id: str,
        v: str,
        name: str,
        file_suffix: str,
        dtransform_id: str,
        storage_id: str,
        instance_id: str,
        time_created: datetime,
        time_updated: datetime,
    ):
        return self.dobject_entity.insert(
            {
                "id": id,
                "v": v,
                "name": name,
                "file_suffix": file_suffix,
                "dtransform_id": dtransform_id,
                "instance_id": instance_id,
                "storage_id": storage_id,
                "owner_id": self.hub.auth.session().user.id.hex,
                "time_created": time_created.strftime("%Y-%m-%d %H:%M:%S"),
                "time_updated": time_updated.strftime("%Y-%m-%d %H:%M:%S"),
            }
        )


class Dtransform:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.dtransform_entity = Entity(hub, "dtransform", ["id", "instance_id"])

    def exists(self, id: str, instance_id: str):
        return self.dtransform_entity.exists({"id": id, "instance_id": instance_id})

    def delete(self, id: str, instance_id: str):
        self.dtransform_entity.delete({"id": id, "instance_id": instance_id})

    def insert(
        self,
        id: str,
        jupynb_id: str,
        jupynb_v: str,
        instance_id: str,
    ):
        return self.dtransform_entity.insert(
            {
                "id": id,
                "jupynb_id": jupynb_id,
                "jupynb_v": jupynb_v,
                "pipeline_run_id": None,
                "instance_id": instance_id,
                "owner_id": self.hub.auth.session().user.id.hex,
            }
        )


class DtransformIn:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.dtransform_in_entity = Entity(
            hub, "dtransform_in", ["dtransform_id", "dobject_id", "dobject_v"]
        )

    def exists(self, dtransform_id: str, dobject_id: str, dobject_v: str):
        return self.dtransform_in_entity.exists(
            {"id": dtransform_id, "dobject_id": dobject_id, "dobject_v": dobject_v}
        )

    def delete(self, dtransform_id: str, dobject_id: str, dobject_v: str):
        return self.dtransform_in_entity.delete(
            {"id": dtransform_id, "dobject_id": dobject_id, "dobject_v": dobject_v}
        )

    def insert(
        self,
        dtransform_id: str,
        dobject_id: str,
        dobject_v: str,
        instance_id: str,
    ):
        return self.dtransform_in_entity.insert(
            {
                "dtransform_id": dtransform_id,
                "dobject_id": dobject_id,
                "dobject_v": dobject_v,
                "instance_id": instance_id,
                "owner_id": self.hub.auth.session().user.id.hex,
            }
        )


class Storage:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.storage_entity = Entity(hub, "storage", ["id"])

    def exists(self, root: Union[Path, CloudPath, str]):
        statement = self.hub.table("storage").select("*").eq("root", str(root))
        data = statement.execute()
        if len(data.data) > 0:
            return True
        return False

    def delete(self, id: str):
        self.storage_entity.delete(
            {
                "id": id,
            }
        )

    def insert(
        self, id: str, root: Union[Path, CloudPath, str], region: str, type: str
    ):
        return self.storage_entity.insert(
            {
                "id": id,
                "root": str(root),
                "region": region,
                "type": type,
            }
        )

    def instance_count(self, storage_id: str):
        statement = self.hub.table("instance").select("*").eq("storage_id", storage_id)
        data = statement.execute()
        return len(data.data)


class Usage:
    def __init__(self, hub: Client) -> None:
        self.hub = hub
        self.usage_entity = Entity(hub, "usage", ["id"])

    def exists(self, id: str):
        return self.usage_entity.exists(
            {
                "id": str(id),
            }
        )

    def delete(self, id: str):
        self.usage_entity.delete(
            {
                "id": str(id),
            }
        )

    def insert(
        self,
        id: str,
        type: str,
        instance_id: str,
        time: datetime,
        dobject_id: str,
        dobject_v: str,
    ):
        return self.usage_entity.insert(
            {
                "id": id,
                "type": type,
                "owner_id": self.hub.auth.session().user.id.hex,
                "instance_id": instance_id,
                "time": str(time),
                "dobject_id": dobject_id,
                "dobject_v": dobject_v,
            }
        )



import sqlmodel as sqm
from lamin_logger import logger
from lndb_setup import settings
from supabase import Client


def get_schema_version():
    assert settings.instance._sqlite_file.exists()
    import lnschema_core  # noqa

    with sqm.Session(settings.instance.db_engine()) as session:
        version = session.exec(sqm.select(lnschema_core.version_yvzi)).all()[-1].v
    return version


def check_schema_version(hub: Client):
    current_version = get_schema_version()
    data = hub.table("version_yvzi").select("*").eq("v", current_version).execute()
    assert len(data.data) == 1
    if data.data[0]["require_upgrade"]:
        logger.error("Your core schema module version is not compatible with the hub.")
        logger.info(
            "Did you already migrate your db to core schema v{current_version}? (y/n)"
        )
        logger.info(
            f"If yes, run `lndb_setup._db.insert.version_yvzi({current_version},"
            " db.settings.user.id)`"
        )
        logger.warning(
            "If no, either migrate your instance db schema to version"
            f" {current_version}.\nOr install the latest version."
        )
        raise RuntimeError("hub needs higher lnschema_core_v")



"""Connect to the hub from LaminDB.

Import the package::

   import lndb_hub

This is the complete API reference:

.. autosummary::
   :toctree: .

   hub
"""

__version__ = "0.5.5"  # denote a pre-release for 0.1.0 with 0.1a1

from ._hub import get_hub_with_authentication, hub  # noqa



# def test_schema_current_version():
#    import lnschema_core  # noqa
#
#    current_version = lnschema_core.__version__
#
#    print(100 * "*", current_version)
#    if current_version == "0.3.1":
#        get_hub_with_authentication()
#
#
# def test_schema_old_version():
#    import lnschema_core  # noqa
#
#    current_version = lnschema_core.__version__
#
#    print(100 * "*", current_version)
#    if current_version == "0.3.0":
#        with pytest.raises(RuntimeError):
#            get_hub_with_authentication()
#
